/*
Copyright (C) 2004  The PARI group.

This file is part of the PARI package.

PARI/GP is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation. It is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY WHATSOEVER.

Check the License for details. You should have received a copy of it, along
with the package; see the file 'COPYING'. If not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include "pari.h"
#include "tree.h"
#include "anal.h"
//ulong   prec, precdl;

const char *Ffuncname[]={"Fseq",
		   "Fmatrix","FmatrixL","FmatrixR",
		   "Faffect",
		   "Ffacteurmat",
		   "Fmatrixelts","Fmatrixlines",
		   "Femptyvec","Femptymat","Fmat",
		   "Flistarg",

		   "Fconst","Fsmall","Fgnil",
		   "Frefarg","Ftag",
		   "Fentry","Fentryfunc","Fdeffunc",

		   /*These nodes are generated by genblock, not by parser*/
		   "Fblock"
};


typedef GEN (*PFGEN)(ANYARG);
#define _ARGS_ argvec[0], argvec[1], argvec[2], argvec[3],\
               argvec[4], argvec[5], argvec[6], argvec[7], argvec[8]

GEN
grandopari(int a)
{
  GEN res;
  long m=1;
  if (tree[a].f!=Fentryfunc || tree[a].x!=OPpow)
    res=evalnode(a);
  else
  {
    int y=tree[a].y;
    res=evalnode(tree[y].x);
    m=itos(evalnode(tree[y].y));
  }
  return ggrando(res,m); 
}

GEN andpari(int a, int b)
{
  if (gcmp0(evalnode(a)))
    return gzero;
  return gcmp0(evalnode(b))?gzero:gun;
}

GEN orpari(int a, int b)
{
  if (!gcmp0(evalnode(a)))
    return gun;
  return gcmp0(evalnode(b))?gzero:gun;
}

GEN gmule(entree *ep, GEN x)
{
  changevalue(ep,gmul(ep->value,x));
  return ep->value;
}

GEN gdive(entree *ep, GEN x)
{
  changevalue(ep,gdiv(ep->value,x));
  return ep->value;
}

GEN gshiftle(entree *ep, long n)
{
  changevalue(ep,gshift(ep->value,n));
  return ep->value;
}

GEN gshiftre(entree *ep, long n)
{
  changevalue(ep,gshift(ep->value,-n));
  return ep->value;
}

GEN gadde(entree *ep, GEN x)
{
  changevalue(ep,gadd(ep->value,x));
  return ep->value;
}

GEN gaddone(entree *ep)
{
  changevalue(ep,gaddgs(ep->value,1));
  return ep->value;
}

GEN gsube(entree *ep, GEN x)
{
  changevalue(ep,gsub(ep->value,x));
  return ep->value;
}

GEN gsubone(entree *ep)
{
  changevalue(ep,gsubgs(ep->value,1));
  return ep->value;
}

GEN gshift_right(GEN x, long n) { return gshift(x,-n); }

extern void changevalue(entree *ep, GEN x);
extern GEN gnil;

char *current_func_name;

void localpari(GEN v)
{
  int i;
  for(i=1;i<lg(v);i++)
    evalnode(v[i]);
}

entree *getentry(int n)
{
  if (tree[n].f!=Fentry) 
    err(talker,"identifier required (found [%s])",Ffuncname[tree[n].f]);
  else
  {
    int x=tree[n].x;
    char *name=(char *)value[x].val.str;
    char *p;
    entree *ep; 
    if (current_func_name)
    {
      p=gpmalloc(strlen(current_func_name)+strlen(name)+2);
      sprintf(p,"%s:%s",current_func_name,name);
    }
    else 
      p=name;
    ep = is_entry(p);
    if (!ep)
      ep=fetch_named_var(p,0);
    if (current_func_name)
      free(p);
    return ep;
  }
  return NULL;/*NOT REACHED*/
}

entree *getentryfunc(int n)
{
  if (tree[n].f!=Fentryfunc) 
    err(talker,"function required (found [%s])",Ffuncname[tree[n].f]);
  else
  {
    int x=tree[n].x;
    char *name=(char *)value[x].val.str;
    entree *ep; 
    ep = is_entry(name);
    if (!ep)
      ep=fetch_named_var(name,0);
    return ep;
  }
  return NULL;/*NOT REACHED*/
}

GEN listtogen(int n, int f)
{
  GEN z;
  int x,i,nb;
  if (n==-1) return cgetg(1,t_VECSMALL);
  for(x=n, i=0; tree[x].f==f ;x=tree[x].x, i++);
  nb=i+1;
  z=cgetg(nb+1,t_VECSMALL);
  for (x=n; i>0; z[i+1]=tree[x].y, x=tree[x].x, i--);
  z[1]=x;
  return z;
}

static GEN
cattovec(int n, int fnum)
{
  int x=n, i=0, nb;
  GEN stack;
  if (n==GNOARG) return cgetg(1,t_VECSMALL);
  while(1)
  {
    int xx=tree[x].x;
    int xy=tree[x].y;
    if (tree[x].f!=Fentryfunc || xx!=fnum) break;
    x=tree[xy].x;
    i++;
  }
  nb=i+1;
  stack=cgetg(nb+1,t_VECSMALL);
  for(x=n;i>0;i--)
  {
    int y=tree[x].y;
    x=tree[y].x;
    stack[i+1]=tree[y].y;
  }
  stack[1]=x;
  return stack;
}

GEN return_data;

GEN evalfunc(int n)
{
  int i,j;
  int x=tree[n].x;
  char *name=(char *)value[x].val.str;
  int ret;
  GEN res;
  char const *p,*q;
  char c;
  PPproto mod;
  entree *ep=tree[n].e;
  void *call; 
  GEN argvec[9];
  long m;
  GEN arg=listtogen(tree[n].y,Flistarg);
  int nb=lg(arg)-1;
  int has_pointer=0;
  GEN gen[9];
  entree *init[9];
  if (!ep)
  {
    ep = is_entry(name);
    if (!ep) 
      err(talker,"unknown function %s",name);
    ep=do_alias(ep);
    tree[n].e=ep;
  }
  if (ep->valence==EpUSER)
  {
    GEN gen[8];
    GEN res;
    GEN arg2=listtogen(tree[(int)ep->args].y,Flistarg);
    long nb2=lg(arg2)-1;
    char *old_func=current_func_name;
    for (i=1;i<=nb;i++)
    {
      current_func_name=ep->name;
      evalnode(arg2[i]);
      current_func_name=old_func;
      gen[i]=gclone(tree[arg2[i]].e->value);
      changevalue(tree[arg2[i]].e,evalnode(arg[i]));
    }
    current_func_name=ep->name;
    for (  ;i<nb2;i++)
    {
      evalnode(arg2[i]);
      gen[i]=gclone(tree[arg2[i]].e->value);
      changevalue(tree[arg2[i]].e,gzero);
    }
    CATCH(returner) {
      CATCH_RELEASE();
      res=gcopy((GEN)return_data);
    } TRY {
      res=evalnode((int)ep->value);
    } ENDCATCH;
    for (i=1;i<=nb2;i++)
    {
      changevalue(tree[arg2[i]].e,gen[i]);
      gunclone(gen[i]);
    }
    current_func_name=old_func;
    return res;
  }
  p=ep->code; call= ep->value;
  if (!call)
    err(impl,name);
  if      (*p < 'a') { ret = RET_GEN; }
  else if (*p == 'v') { ret = RET_VOID; p++; }
  else if (*p == 'i') { ret = RET_INT;  p++; }
  else if (*p == 'l') { ret = RET_LONG; p++; }
  else ret = RET_GEN;
  i=0;j=1;
  while(*p=='G')
  {
    argvec[i++]=evalnode(arg[j++]);
    p++;
  }
  q=p; 
  while((mod=parseproto(&p,&c)))
  {
    if (j<=nb && arg[j]!=GNOARG && mod!=PPstar)
      mod=PPstd;
    switch(mod)
    {
    case PPstd:
      switch(c)
      {
      case 'G':
        argvec[i]=evalnode(arg[j++]);
        break;
      case 'p':
        argvec[i]=(GEN)prec;
        break;
      case 'M':
      case 'L':
        argvec[i]=(GEN)itos(evalnode(arg[j++]));
        break;
      case 'n':
        argvec[i]=(GEN)varn(evalnode(arg[j++]));
        break;
      case '&':
        {
          int a=arg[j++];
          if (tree[a].f!=Frefarg)
            err(talker,"& expected");
          a=tree[a].x;
          if (!tree[a].e) tree[a].e=getentry(a);
          has_pointer |= (1 << i);
          init[i] = tree[a].e;
          gen[i]=tree[a].e->value;
          argvec[i]=(GEN) &gen[i];
        }
        break;
      case 'P':
        argvec[i]=(GEN)precdl;
        break; 
      case 'I':
      case 'E':
        argvec[i]=(GEN)arg[j++];
        break;
      case 'S':
      case 'V':
        if (!tree[arg[j]].e) tree[arg[j]].e=getentry(arg[j]);
        argvec[i]=(GEN)tree[arg[j++]].e;
        break;
      case 's':
        argvec[i]=(GEN)GENtostr(evalnode(arg[j++]));
        break;
      case 'r':
        {
          int a=arg[j++];
          if (tree[a].f==Fentry)
            argvec[i]=(GEN)value[tree[a].x].val.str;
          else
            argvec[i]=(GEN)GENtostr(evalnode(a));
        }
        break;
      default:
        err(talker,"Unknown prototype code `%c' for `%s'",c,name);
      }
      break;
    case PPdefault:
      j++;
      switch(c)
      {
      case 'G':
      case '&':
      case 'V':
      case 'r':
        argvec[i]=(GEN)NULL;
        break;
      case 'n':
        argvec[i]=(GEN)-1;
        break;
      case 'E':
      case 'I':
        argvec[i]=(GEN)GNOARG;
        break;
      default:
        err(talker,"Unknown default in prototype code `%c' for `%s'",c,name);
      }
      break;
    case PPdefaultmulti:
      j++;
      switch(c)
      {
      case 'L':
      case 'M':
        argvec[i]=(GEN)strtol(q+1,NULL,10);
        break;
      case 's':
      case 'r':
        argvec[i]=(GEN)"";
        break;
      default:
        err(talker,"Unknown default in prototype code `%c' for `%s'",c,name);
      }
      break;
    case PPstar:
      switch(c)
      {
      case 'V':
        {
          int x=tree[arg[j]].x;
          int y=tree[arg[j]].y;
          if (tree[arg[j]].f!=Faffect)
            err(talker,"= expected");
          if (!tree[x].e) tree[x].e=getentry(x);
          argvec[i]=(GEN)tree[x].e;
          arg[j]=y;
        }
        break;
      case 'I':
        argvec[i]=(GEN)arg;
        break;
      case 's':
        {
          int n=nb+1-j;
          int k,l,l1,m;
          GEN r,g=cgetg(n+1,t_VEC);
          for(l1=0,k=1;k<=n;k++)
          {
            g[k]=(long)cattovec(arg[j+k-1],OPcat);
            l1+=lg(g[k])-1;
          }
          r=cgetg(l1+1,t_VEC);
          for(m=1,k=1;k<=n;k++)
          {
            for(l=1;l<lg(g[k]);l++)
              r[m++]=(long)evalnode(mael(g,k,l));
          }
          argvec[i]=r;
        }
        break;
      default:
        err(talker,"Unknown prototype code `%c*' for `%s'",c,name);
      }
      break;
    default:
      err(bugparier,"PPproto %d in gencallfunc",mod);
    }
    i++;
    q=p;
  }
  for (  ;i<9;i++) argvec[i]=NULL;
  switch (ret)
  {
  case RET_GEN: 
    res = ((PFGEN)call)(_ARGS_);
    break;

  case RET_INT:
    m = (long)((int (*)(ANYARG))call)(_ARGS_);
    res = stoi(m); break;

  case RET_LONG:
    m = ((long (*)(ANYARG))call)(_ARGS_);
    res = stoi(m); break;

  case RET_VOID:
    ((void (*)(ANYARG))call)(_ARGS_);
    res = gnil; break;
  default:
    res=gnil;
  }
  if (has_pointer)
    for(i=0;i<8;i++)
      if (has_pointer&(1<<i))
        changevalue(init[i],gen[i]);
  return res;
}

GEN 
evalfacteurmat(int n)
{ 
  int x=tree[n].x;
  int y=tree[n].y;
  int yx=tree[y].x;
  int yy=tree[y].y;
  GEN gx=evalnode(x);
  long t=typ(gx);
  if (yy==-1)
  {
    long b=itos(evalnode(yx));
    if(t==t_VECSMALL)
      return stoi(gx[b]);
    if(t==t_VEC || t==t_COL || t_MAT)
      return (GEN) gx[b];
  }
  else if (t==t_MAT)
  {
    long b=itos(evalnode(yx));
    long c=itos(evalnode(yy));
    return gcoeff(gx,b,c);
  }
  err(talker,"object has no components");
  return NULL; /* NOT REACHED */ 
}

GEN evalvector(int n)
{
  int x=tree[n].x;
  int i;
  GEN ret;
  GEN arg=listtogen(x,Fmatrixelts);
  long l=lg(arg);
  ret=cgetg(l,t_VEC); 
  for(i=1;i<l;i++)
    ret[i]=(long)evalnode(arg[i]);
  return ret; 
}
#define STACKSZ 1024
GEN evalmatrix(int n)
{
  int line[STACKSZ];
  int arg[STACKSZ];
  int x=tree[n].x;
  int i,j,nb=0,nbline,nbcol=0;
  GEN ret;
  nbline=listtostack(x,Fmatrixlines,line,STACKSZ,"Matrix too long.",n);
  for(i=0;i<nbline;i++)
  {
    int k;
    k=listtostack(line[i],Fmatrixelts,arg+nb,STACKSZ-nb,"Matrix too long.",n);
    if (i && k!=nbcol)
      err(talker,"Matrix must be rectangular");
    nbcol=k;
    nb+=k;
  }
  ret=cgetg(nbcol+1,t_MAT);
  for(i=1;i<=nbcol;i++)
  {
    GEN v=cgetg(nbline+1,t_COL);
    for(j=1; j<=nbline; j++)
      v[j]=(long) evalnode(arg[(i-1)+(j-1)*nbcol]);
    ret[i]=(long) v;
  }
  return ret;
}

GEN evalnode(int n)
{
  int x,y;
  if (n<0)
  {
    if (n==GNOARG)
      err(talker,"missing mandatory argument");
    else
      err(bugparier,"n<0 in evalnode");
  }
  if (DEBUGLEVEL)
    fprintferr("[%d]:%s\n",n,Ffuncname[tree[n].f]);
  x=tree[n].x;
  y=tree[n].y;
  switch(tree[n].f)
  {
  case Fseq:
    evalnode(x);
    return evalnode(y);
  case Fmatrix:
  case FmatrixL:
  case FmatrixR:
    break;
  case Ffacteurmat:
    return evalfacteurmat(n);
  case Faffect:
    {
      GEN g=evalnode(y);
      if (!tree[x].e) tree[x].e=getentry(x); 
      changevalue(tree[x].e,g);
      return g;
    }
  case Fconst:
    switch(value[x].type)
    {
    case CSTsmall:
      return stoi(value[x].val.small);
    case CSTsmallreal:
      return dbltor(value[x].val.small*1.);
    case CSTgen: 
      return value[x].val.gen;
    case CSTdot: 
      return itor(value[x].val.gen,prec);
    case CSTstr:
      return STRtoGENstr((char *)value[x].val.str);
    case CSTquote:
      return polx[fetch_user_var((char *)value[x].val.str)];
    }
    break;
  case Fsmall:
    return stoi(x);
  case Fmatrixelts:
  case Fmatrixlines:
    break;
  case Femptyvec:
    return cgetg(1,t_VEC);
  case Femptymat:
    return cgetg(1,t_MAT);
  case Fmat:
    if (tree[x].f==Fmatrixlines)
      /*if there is at least 2 lines it is a matrix*/
      return evalmatrix(n);
    else
      /*else it is a vector*/
      return evalvector(n);
  case Flistarg:
    break;
  case Frefarg:
    err(talker,"unexpected &"); 
    break;
  case Fentry:
    if (!tree[n].e)
    {
      char *name=(char *)value[x].val.str;
      entree *ep=is_entry(name);
      if (ep && EpVALENCE(ep)!=EpVAR && EpVALENCE(ep)!=EpGVAR)
      {
        tree[n].f=Fentryfunc;
        tree[n].e=ep;
        return evalfunc(n);
      }
      tree[n].e=getentry(n);
    }
    return tree[n].e->value;
  case Fentryfunc:
    return evalfunc(n);
  case Fdeffunc:
    {
      entree *ep = getentryfunc(x);
      ep->valence=EpUSER;
      ep->value = (void *) y;
      ep->args = (void *) x;
      return gnil;
    }
  case Fblock:
    break;
  case Fgnil:
    return gnil;
  case Ftag:
    return evalnode(x);    
  }
  err(talker,"node %s not implemented yet\n",Ffuncname[tree[n].f]);
  return gnil;
}

